#include "main.h"
#include <cassert>
#include <cstdio>
#include <iostream>
#include <string>
#include <vector>

#include <QApplication>
#include <QRegExp>
#include <QtWebKit>

#include <termios.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

using namespace std;

//
//  MARKUS URL AND COURSE NAME
//
QString markus_base_url("https://markus002.student.cs.uwaterloo.ca/markus_cs135_f/");
QString course("CS135");

//
//  YOU CAN ENTER YOUR MARKUS LOGIN INFORMATION HERE
//  You can also leave them blank. You'll be prompted for the username and/or password.
//
#define USERNAME ""
#define PASSWORD ""


enum State { DoNothing, LoggingIn, GoToSubmissionsPage, OnSubmissionsPage,Marking } state = LoggingIn;
vector <QString> newmarks;
const int waittime = 1300; // in milliseconds

// determines if s is a valid Quest ID
bool isQuestID(QString s) {
    if(s.endsWith("-student")) return true;

    if(s.size() <= 8) {
        // length is OK, check characters
        for(int i = 0; i < s.size(); i++) {
            if(! (s[i].isLower() || s[i].isNumber()))
                return false;
        }
        return true;
    }
    return false;
}


// prints a QString to os
ostream& operator<<(ostream &os, const QString &str)
{
    return os << str.toStdString();
}


//
// Implement Crawler class's functions
//

Crawler::Crawler(int urlid, QString r, QWidget *parent) : QWidget(parent)
{
    accesstype = UrlId;
    regex = r;
    output = true;
    initialize(); 

    // create URL we should go to
    QString url = QString("%1en/main/results/edit/%2").arg(markus_base_url).arg(urlid);

    if(output) cout << "Going to " << url.toStdString() << endl;
    view->load(QUrl(url));
    view->show();
}

Crawler::Crawler(QString a, QString qid, QWidget *parent) : QWidget(parent)
{
    accesstype = Assn;
    assn = a; questid = qid;
    output = true;
    initialize();

    QString url = QString("%1en/main/assignments").arg(markus_base_url);

    if(output) cout << "Going to " << url.toStdString() << endl;
    view->load(QUrl(url));
    view->show();
}

Crawler::Crawler(QString a, QWidget *parent) : QWidget(parent)
{
    accesstype = Urls;
    assn = a;
    output = false;
    initialize();

    QString url = QString("%1en/main/assignments").arg(markus_base_url);

    if(output) cout << "Going to " << url.toStdString() << endl;
    view->load(QUrl(url));
    view->show();
}

void Crawler::collectURLs()
{
    if(output) cout << "Collecting URLs..." << endl;

    QString markusBaseURL("https://markus002.student.cs.uwaterloo.ca/");

    QWebElementCollection elems(frame->documentElement().findAll("TD A"));
    foreach (QWebElement elem, elems) {
        if(isQuestID(elem.toPlainText())) {
            cout << elem.toPlainText() << "," << markusBaseURL + elem.attribute("href") << endl;

            id2url[elem.toPlainText()] = markusBaseURL + elem.attribute("href"); 
        } 
    }

    if(output) cout << "Found " << id2url.size() << " students." << endl;
}

//  clicks the HTML with id s
void Crawler::clickjsid(QString s)
{
    QString javascriptCode("var evObj = document.createEvent('MouseEvents');"
                           "evObj.initEvent( 'click', true, true );");
    javascriptCode += "document.getElementById('" + s + "').dispatchEvent(evObj);";
    frame->documentElement().evaluateJavaScript(javascriptCode);
}

void Crawler::checkmarks()
{
    unsigned int criteria_marked = 0;

    QWebElementCollection elems(frame->documentElement().findAll("LI"));
    foreach (QWebElement elem, elems) {
        if(elem.attribute("class").startsWith("mark_criterion_holder_"))
        {
            QString criteriaName = elem.findFirst("STRONG").toPlainText();
            QString currentMark = elem.findFirst("DIV[class=\"padding_left\"]").toPlainText();
            cout << "Found the criteria: " << criteriaName << endl;
            cout << "\tMark: " << currentMark << endl;
            
            if(criteria_marked < newmarks.size()) {
                bool conversionok;
                int newmark = newmarks[criteria_marked].toInt(&conversionok);
                if(conversionok && 0 <= newmark && newmark <= 4
                   && !currentMark.contains(newmarks[criteria_marked] + ".0: "))
                {
                    cout << "Found a mark that doesn't agree..." << endl;
                    cout << "Setting marks again..." << endl;
                    loadFinished_handler(true);
                    return;
                }
                criteria_marked++;
                if(criteria_marked >= newmarks.size()) break;
            }
        }
    }

    // was able to assign marks successfully
    if(newmarks.size() > 0)
        qApp->quit();
    else
        state = DoNothing;
}

//
//   PUBLIC SLOTS
//

void Crawler::loadStarted_handler()
{
    if(output) cout << "Started loading page..." << endl;
}

void Crawler::loadProgress_handler(int progress)
{
    if(output) cout << "Progress: " << progress << endl;
}

void Crawler::loadFinished_handler(bool b)
{
    if(output) cout << "Finished loading with value: " << b << endl;

    if(state == LoggingIn) {
        if(output) cout << "Logging in..." << endl; 
        // Log in to MarkUs through CAS
        QWebElement elem = frame->documentElement();
        elem.evaluateJavaScript( QString("document.getElementById('username').value='%1'").arg(username) );
        elem.evaluateJavaScript( QString("document.getElementById('password').value='%1'").arg(password) );
        elem.evaluateJavaScript("document.forms['fm1'].submit.click()");
        if(output) cout << "Finished logging in..." << endl;
        if(accesstype == UrlId)
            state = Marking;
        else
            state = GoToSubmissionsPage;
    }
    else if(state == Marking) {
        // check if regex string was found
        QString htmlcontents(frame->toHtml());
        if(!htmlcontents.contains(QRegExp(regex))) {
            if(output) {
                cout << "----------------------------------------------------------------" << endl;
                cout << frame->toHtml() << endl;
                cout << "----------------------------------------------------------------" << endl;
                cout << "Regular expression " << regex << " not found." << endl;
                cout << "HTML code is above." << endl;
                cout << "Quitting..." << endl;
            }
            qApp->quit();
        }
        else {
            unsigned int criteria_marked = 0;

            // The regex was found. Find and display all criteria.
            QWebElementCollection elems(frame->documentElement().findAll("LI"));
            foreach (QWebElement elem, elems) {
                if(criteria_marked < newmarks.size()
                   && elem.attribute("class").startsWith("mark_criterion_holder_"))
                {
                    QString criteriaID = elem.findFirst("TABLE").attribute("ID");
                    bool conversionok;
                    int newmark = newmarks[criteria_marked].toInt(&conversionok);
                    if(conversionok && 0 <= newmark && newmark <= 4)
                        clickjsid(criteriaID.replace("table", newmarks[criteria_marked]));
                    criteria_marked++;
                } // end if
            } // end foreach
            QTimer::singleShot(waittime, this, SLOT(checkmarks()));
        } // end else
    } // end else if
    else if(state == GoToSubmissionsPage) {
        // find the assignment link.
        int order_in_list = 1;
        QWebElementCollection elems(frame->documentElement().findAll("A[class=assignment_list]"));
        foreach (QWebElement elem, elems) {
            if(elem.toPlainText().contains(QRegExp(assn))) {
                if(output) cout << "Marking " << elem.toPlainText() << endl;
                QString submission_url
                  = QString("%1en/main/submissions/browse/%2?desc=&sort_by=group_name&filter=none&per_page=1000").arg(markus_base_url).arg(order_in_list);
                view->load(QUrl(submission_url));
                state = OnSubmissionsPage;
                return;
            }
            order_in_list++;
        } // end foreach
        cout << "Found no assignment called " << assn << endl;
        state = DoNothing;
    } // end else if
    else if(state == OnSubmissionsPage) {
        collectURLs();
        if(accesstype == Assn) {
            view->load(QUrl(id2url[questid]));
            state = Marking;
        } else if(accesstype == Urls) state = DoNothing;
        else state = DoNothing;
    } // end else if
} // end function definition

void Crawler::sslErrors_handler(QNetworkReply * reply, const QList<QSslError> & errors)
{
    if(output) cout << "Got some SSL errors..." << endl;
    reply->ignoreSslErrors();
}

void Crawler::errorQuit()
{
    cerr << "Cannot set score. Quitting" << endl;
    qApp->exit(1);
}

// initializes some of the variables in Crawler class
void Crawler::initialize()
{
    // Important objects
    view = new QWebView();
    page = view->page();
    frame = page->mainFrame();
    network = page->networkAccessManager();
   
    // make connections
    QObject::connect(view, SIGNAL(loadStarted()), this, SLOT(loadStarted_handler()));
    QObject::connect(view, SIGNAL(loadProgress(int)), this, SLOT(loadProgress_handler(int)));
    QObject::connect(view, SIGNAL(loadFinished(bool)), this, SLOT(loadFinished_handler(bool)));
    QObject::connect(network, SIGNAL(sslErrors(QNetworkReply *, const QList<QSslError> &)),
                     this, SLOT(sslErrors_handler(QNetworkReply *, const QList<QSslError> &)));
    
    // get username and password
    #ifdef USERNAME
        username = USERNAME;
    #endif

    #ifdef PASSWORD
        password = PASSWORD;
    #endif
    
    string input;
    if(username.isEmpty()) {
        if(output) cout << "Enter Quest ID: ";
        cin >> input;
        username = QString::fromStdString(input);
    }
    if(password.isEmpty()) {
        struct termios orig;
        tcgetattr(STDIN_FILENO, &orig);

        struct termios newt = orig;
        newt.c_lflag &= ~ECHO;
        bool inTerminal = tcsetattr(STDIN_FILENO, TCSANOW, &newt) + 1;

        if(output) cout << "Enter Quest password for " << username << ": ";
        cin >> input;
        password = QString::fromStdString(input);

        if(inTerminal) tcsetattr(STDIN_FILENO, TCSANOW, &orig);
 
    }

    // quite after 1 minute
    QTimer::singleShot(60000, this, SLOT(errorQuit()));
}

//
//  MAIN FUNCTION
//

int main(int argc, char *argv[])
{
    QApplication app(argc, argv);

    // Make sure the Markus URL ends with a /
    // If it does not, add it.
    if(! markus_base_url.endsWith('/'))
        markus_base_url.append('/');

    // Check which format to use
    if(argc >= 5 && QString(argv[1]) == "-urlid") { 
        int urlid = QString(argv[2]).toInt();
        QString regex = QString("%1 %2 : %3").arg(course).arg(argv[3]).arg(argv[4]);
        Crawler c(urlid, regex);
        for(int i = 5; i < argc; ++i)
            newmarks.push_back(QString(argv[i]));
        return app.exec();
    }
    else if(argc >= 4 && QString(argv[1]) == "-assn") {
        Crawler c(argv[2], argv[3]);
        for(int i = 4; i < argc; ++i)
            newmarks.push_back(QString(argv[i]));
        return app.exec();
    }
    else if(argc == 3 && QString(argv[1]) == "-urls") {
        Crawler c(argv[2]);
        return app.exec();
    }
    else {
        cout << "Format 1: ./markus -urlid 4244 a06 yc2lee-student m1 m2 ... mN" << endl;
        cout << "Format 2: ./markus -assn 'Assignment 07' yc2lee-student m1 m2 ... mN" << endl;
        cout << "Format 3: ./markus -urls 'Assignment 07'" << endl;
        cout << "The student will get the marks m1, m2, ... mN. If no m's are given," << endl;
        cout << "then this program will only display marks, not change them." << endl;
        return 1;
    }
}
